/**
 * Copyright (c) 2011 Chase
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

package cs.geom;

import java.awt.geom.Line2D;
import java.util.Formatter;

public class Line {
	/** The first vector of this line */
	public Vector a;
	
	/** The second vector of this line */
	public Vector b;
	
	/**
	 * Creates a new line with two vectors at (0,0)
	 */
	public Line() {
		this.a = new Vector();
		this.b = new Vector();
	}
	
	/**
	 * Creates a new line from the given vectors.
	 */
	public Line(Vector p, Vector q) {
		this.a = p;
		this.b = q;
	}
	
	/**
	 * Returns the length of this line.
	 */
	public final double length() {
		return a.distance(b);
	}
	
	/**
	 * Returns the center of this line.
	 */
	public final Vector getCenter() {
		Vector vec = Vector.add(a, b);
		vec.x /= 2.0;
		vec.y /= 2.0;
		return vec;
	}
	
	/**
	 * Returns the directional unit vector associated with this line.
	 * @return
	 */
	public final Vector getDirection() {
		Vector dir = Vector.sub(this.b, this.a);
		dir.normalize();
		return dir;
	}
	
	/**
	 * Returns true if this line segment and the given line segment intersect.
	 */
	public final boolean intersects(Line q) {
		/* dom is chase speak for denominator */
		double dom = (q.b.y-q.a.y)*(this.b.x-this.a.x)
					-(q.b.x-q.a.x)*(this.b.y-this.a.y);
		double ua = (q.b.x-q.a.x)*(this.a.y-q.a.y)
					-(q.b.y-q.a.y)*(this.a.x-q.a.x);
		double ub = (this.b.x-this.a.x)*(this.a.y-q.a.y)
					-(this.b.y-this.a.y)*(this.a.x-q.a.x);
		
		/* If dom is 0 then they are parallel */
		/* If ua and ub are also 0, then they are coincident */
		if(Math.abs(dom) < 1e-6)
			return false;
		
		ua /= dom;
		ub /= dom;
		if(ua >= 0.0 && ua <= 1.0 && ub >= 0.0 && ub <= 1.0) {
			return true;
		}
		return false;
	}
	
	/**
	 * Returns the nearest squared distance to the given Vector on this Line
	 */
	public final double ptLineDistSq(Vector p) {
		return Line2D.ptLineDistSq(a.x, a.y, b.x, b.y, p.x, p.y);
	}
	
	/**
	 * Returns the nearest distance to the given Vector on this Line
	 */
	public final double ptLineDist(Vector p) {
		return Line2D.ptLineDist(a.x, a.y, b.x, b.y, p.x, p.y);
	}
	
	/**
	 * Returns the nearest squared distance to the given Vector on this Line Segment
	 */
	public final double ptSegDistSq(Vector p) {
		return Line2D.ptSegDistSq(a.x, a.y, b.x, b.y, p.x, p.y);
	}
	
	/**
	 * Returns the nearest distance to the given Vector on this Line Segment
	 */
	public final double ptSegDist(Vector p) {
		return Line2D.ptSegDist(a.x, a.y, b.x, b.y, p.x, p.y);
	}
	
	/**
	 * Converts this to a string summary
	 */
	public String toString() {
		return (new Formatter()).format("Line[%.4f,%.4f,%.4f,%.4f]", a.x, a.y, b.x, b.y).toString();
	}
}
