/**
 * Copyright (c) 2011 Chase
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

package cs.geom;

import java.util.Formatter;

/**
 * Defines a rectangular shape.
 */
public class Rectangle {
	/** The minimal corner of this rectangle */
	public Vector min;
	/** The maximal corner of this rectangle */
	public Vector max;
	/** The size of this rectangle, for performance reasons */
	public Vector size;

	/**
	 * Creates an empty rectangle.
	 */
	public Rectangle() {
		min = new Vector();
		max = min.copy();
		size = min.copy();
	}

	/**
	 * Creates a rectangle based off the min and the size.
	 */
	public Rectangle(double x, double y, double w, double h) {
		min = new Vector(x,y);
		max = new Vector(x+w, y+h);
		size = new Vector(w, h);
	}

	/**
	 * Creates a rectangle from its center with the given width and height.
	 */
	public Rectangle(Vector center, double w, double h) {
		min = new Vector();
		max = min.copy();
		size = min.copy();

		set(center,w,h);
	}

	/**
	 * Sets this rectangle according to the given
	 */
	public final void set(Vector center, double w, double h) {
		min.set(center);
		max.set(center);
		size.set(w,h);

		min.x -= size.x / 2.0;
		min.y -= size.y / 2.0;

		max.x += size.x / 2.0;
		max.y += size.y / 2.0;
	}

	/**
	 * Returns true if this rectangle contains the given vector.
	 */
	public final boolean contains(Vector a) {
		if(a.x > min.x && a.x < max.x
				&& a.y > min.y && a.y < max.y) {
			return true;
		}
		return false;
	}

	/**
	 * Returns true if this rectangle contains the given vector.
	 */
	public final boolean contains(double x, double y) {
		if(x > min.x && x < max.x
				&& y > min.y && y < max.y) {
			return true;
		}
		return false;
	}

	/**
	 * Returns true if this rectangle intersects the given rectangle.
	 */
	public final boolean intersects(Rectangle r) {
		if(size.x <= 0 || size.y <= 0 || r.size.x <= 0 || r.size.y <= 0) {
			return false;
		}
		return (r.max.x > min.x &&
				r.max.y > min.y &&
				r.min.x < max.x &&
				r.min.y < max.y);
	}


	/**
	 * Intersects Line, stolen from Rectangle2D
	 */
	public final boolean intersects(Line n) {
		/* Check if either point is within this. */
		if(contains(n.a) || contains(n.b))
			return true;
		/* Check each of the edge lines for intersection */
		for(Line line : getLines()) {
			if(n.intersects(line))
				return true;
		}
		/* If all that failed, the chance of intersection is very low. */
		return false;
	}

	/**
	 * Returns an array of 4 vectors, which represent the corners of this rectangle.
	 */
	public final Vector[] getCorners() {
		return new Vector[] {
				min.copy(),
				new Vector(max.x, min.y),
				new Vector(min.x, max.y),
				max.copy()
		};
	}

	public final Vector[] getFasterCorners() {
		return new Vector[] {
				min,
				new Vector(max.x, min.y),
				new Vector(min.x, max.y),
				max
		};
	}

	/**
	 * Returns the center of this rectangle.
	 */
	public final Vector getCenter() {
		return new Vector(min.x + size.x / 2.0, min.y + size.y / 2.0);
	}

	/**
	 * Returns an array of 4 lines, which constitute the edges of this rectangle.
	 * @return
	 */
	public final Line[] getLines() {
		return new Line[] {
				new Line(min.copy(), new Vector(max.x, min.y)),
				new Line(min.copy(), new Vector(min.x, max.y)),
				new Line(new Vector(min.x, max.y), max.copy()),
				new Line(new Vector(max.x, min.y), max.copy())
		};
	}

	/**
	 * Converts this to a string summary
	 */
	@Override
	public String toString() {
		return (new Formatter()).format("Rectangle[%.4f,%.4f,%.4f,%.4f]", min.x, min.y, size.x, size.y).toString();
	}

	public java.awt.geom.Rectangle2D toRectangle2D() {
		return new java.awt.geom.Rectangle2D.Double(min.x,min.y,size.x,size.y);
	}
}
