/**
 * Copyright (c) 2011 Chase
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */

package cs.geom;

import java.util.Formatter;

/**
 * A basic point/vector class, to replace the need for Point2D
 */
public class Vector {
	
	/** The x location */
	public double x;
	
	/** The y location */
	public double y;
	
	/**
	 * Constructs a vector at (0,0)
	 */
	public Vector() {
		this.x = this.y = 0;
	}
	
	/**
	 * Constructs a vector.
	 */
	public Vector(double x1, double y1) {
		this.x = x1;
		this.y = y1;
	}
	
	/** POINT METHODS */
	
	/**
	 * Sets this vector to be equal to the given vector.
	 */
	public final void set(Vector vec) {
		this.x = vec.x;
		this.y = vec.y;
	}
	
	/**
	 * Sets this vector to be equal to the given coordinates.
	 */
	public final void set(double x, double y) {
		this.x = x;
		this.y = y;
	}
	
	/**
	 * Sets this vector to be equal to the given projection. Equivalent to calling set then project.
	 */
	public final void setProject(Vector vec, double angle, double distance) {
		this.x = vec.x + Math.sin(angle) * distance;
		this.y = vec.y + Math.cos(angle) * distance;
	}
	
	/**
	 * Projects this point, by an angle and distance
	 */
	public final void project(double angle, double distance) {
		this.x += Math.sin(angle) * distance;
		this.y += Math.cos(angle) * distance;
	}
	
	/**
	 * Projects from this point, by an angle and distance
	 */
	public final Vector projectNew(double angle, double distance) {
		return project(this,angle,distance);
	}
	
	/**
	 * Projects from the vector along the angle, at the a distance
	 */
	public static final Vector project(Vector v, double angle, double distance) {
		return new Vector(
				v.x + Math.sin(angle) * distance,
				v.y + Math.cos(angle) * distance
			);
	}
	
	/**
	 * Creates a vector based on a given angle and distance as its arguments
	 */
	public static final Vector create(double angle, double distance) {
		return new Vector(
				Math.sin(angle) * distance,
				Math.cos(angle) * distance
			);
	}
	
	/**
	 * Calculates the angle from this point to point p.
	 */
	public final double angleTo(Vector p) {
		return Math.atan2(p.x - this.x, p.y - this.y);
	}
	
	/**
	 * Calculates the angle from point p to this point.
	 */
	public final double angleFrom(Vector p) {
		return Math.atan2(this.x - p.x, this.y - p.y);
	}
	
	/**
	 * Calculates the distance from this point to the point p
	 */
	public final double distance(Vector p) {
		return distance(p.x,p.y,this.x,this.y);
	}
	
	/**
	 * Calculates the squared distance from this point to the point p
	 */
	public final double distanceSq(Vector p) {
		return distanceSq(p.x,p.y,this.x,this.y);
	}
	
	/**
	 * Calculates the Manhattan distance from this point to the point p
	 */
	public final double distanceTaxi(Vector p) {
		return Math.abs(p.x-x)+Math.abs(p.y-y);
	}
	
	/**
	 * Calculates the distance between two points (x1,y1) and (x2,y2)
	 */
	public static final double distance(double x1, double y1, double x2, double y2) {
		return Math.sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
	}
	
	/**
	 * Calculates the squared distance between two points (x1,y1) and (x2,y2)
	 */
	public static final double distanceSq(double x1, double y1, double x2, double y2) {
		return (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1);
	}
	
	/** VECTOR METHODS */
	
	/**
	 * Returns the length of this vector, equivalent to taking the distance to zero. 
	 */
	public final double length() {
		return Math.sqrt(this.x*this.x + this.y*this.y);
	}
	
	/**
	 * Returns the squared length of this vector, equivalent to taking the squared distance to zero. 
	 */
	public final double lengthSq() {
		return this.x*this.x + this.y*this.y;
	}
	
	/**
	 * Normalizes this vector, thus turning it into a unit vector. Equivalent to dividing
	 * each dimension by the vectors length.
	 */
	public final void normalize() {
		double length = length();
		this.x /= length;
		this.y /= length;
	}
	
	/**
	 * this = (this + p) / 2.0
	 */
	public final void average(Vector p) {
		this.x = (this.x + p.x) / 2.0;
		this.y = (this.y + p.y) / 2.0;
	}
	
	/**
	 * Calculates the dot product of this and the given vector.
	 */
	public final double dot(Vector p) {
		return this.x * p.x + this.y * p.y;
	}
	
	/**
	 * Scales this vector as such that x = x * scale and y = y * scale.
	 */
	public final void scale(double scale) {
		x *= scale;
		y *= scale;
	}
	
	public static final Vector scale(Vector v, double scale) {
		return new Vector(
				v.x *= scale,
				v.y *= scale
			);
	}
	
	/**
	 * Subtracts vector b from this vector.
	 */
	public final void sub(Vector b) {
		this.x -= b.x;
		this.y -= b.y;
	}
	
	/**
	 * Adds vector b to this vector.
	 */
	public final void add(Vector b) {
		this.x += b.x;
		this.y += b.y;
	}
	
	/**
	 * Subtracts x,y from this vector.
	 */
	public final void sub(double x1, double y1) {
		this.x -= x1;
		this.y -= y1;
	}
	
	/**
	 * Adds x,y to this vector.
	 */
	public final void add(double x1, double y1) {
		this.x += x1;
		this.y += y1;
	}
	
	/**
	 * Converts this vector to an angle
	 */
	public final double toAngle() {
		return Math.atan2(this.x, this.y);
	}
	
	/**
	 * Rotates this vector by the given angle.
	 */
	public final void rotate(double angle) {
		double length = length();
		double currentAngle = toAngle();
		x = 0; y = 0;
		project(currentAngle + angle, length);
	}
	
	/**
	 * Rotates this vector by the given angle around the given center.
	 */
	public final void rotate(Vector center, double angle) {
		sub(center);
		rotate(angle);
		add(center);
	}

	/**
	 * Returns a vector c = a - b. As such that
	 * c.x = a.x - b.x and c.y = a.y - b.y
	 */
	public final static Vector sub(Vector a, Vector b) {
		return new Vector(a.x - b.x, a.y - b.y);
	}
	
	/**
	 * Returns a vector c = a + b. As such that
	 * c.x = a.x + b.x and c.y = a.y + b.y
	 */
	public final static Vector add(Vector a, Vector b) {
		return new Vector(a.x + b.x, a.y + b.y);
	}
	
	/**
	 * Returns a copy of this vector.
	 */
	public final Vector copy() {
		return new Vector(this.x, this.y);
	}
	
	public int hashCode() {
		return new Double(x).hashCode() ^ new Double(y).hashCode();
	}
	
	public boolean equals(Object v) {
		if(v == this)
			return true;
		if(v instanceof Vector) {
			if(((Vector)v).x == this.x
			&& ((Vector)v).y == this.y)
				return true;
			return false;
		}
		return false;
	}
	
	/**
	 * Converts this to a string summary
	 */
	public String toString() {
		return (new Formatter()).format("Vector[%.4f,%.4f]", x, y).toString();
	}
}