/**
 * Copyright (c) 2011 Chase
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 *    1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 
 *    2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 
 *    3. This notice may not be removed or altered from any source
 *    distribution.
 */
/*
 * Dear Future Chase,
 * 
 * I cannot even begin to express how deeply sorry I am.
 * But I assure you, at the time, I (probably) knew what I was doing.
 * 
 * -Chase
 * 
 * P.S. I am trying to leave comments,
 *      but you know how bad I am at those.
 */

package cs.move;

import java.awt.Color;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import cs.Chikaku;
import cs.geom.*;
import cs.utils.*;

import ags.utils.KdTree;
import ags.utils.KdTree.Entry;
import robocode.*;
import robocode.util.Utils;

/**
 * Movement
 * <br/><br/>
 * In Japanese, Kakeru means "Dash" (in some cases)
 * @author Chase
 */
public abstract class Kakeru extends Chikaku {
	/**
	 * Keep in mind with heat waves we will have waves 2 rounds earlier
	 * the the gun heat would otherwise indicate
	 */
	private static final int TimeBeforeFirstWaveToStartTurning = 4;
	private static final int MaxTurnHistoryLength = 32;
	private static final int MaxHeatWaveQueueLength = 4;
	private static final int AverageSingleRoundTime = 1200;

	/*
	 * Distancer Constants, control how the distancer acts,
	 * limiter is linear equation
	 */
	private static final double distBestDistance = 500.0;

	/* distancer angle limiting */
	private static final double distMinDistance = 300;
	private static final double distMinRotation = Math.PI/4.0;
	private static final double distMaxDistance = 800;
	private static final double distMaxRotation = Math.PI/10.0;
	private static final double distMultiplier = (distMaxRotation-distMinRotation)/(distMaxDistance-distMinDistance);
	private static final double distAdder = distMinRotation-distMultiplier*distMinDistance;

	private static final MaximumDeque<Double> enemyBulletPower;
	private static final KdTree<Data> tree;

	static {
		enemyBulletPower = new MaximumDeque<Double>(MaxHeatWaveQueueLength);
		enemyBulletPower.addLast(3.0);

		int dimensions = new Data().getDataArray().length;
		tree = new KdTree.SqrEuclid<Data>(dimensions, 0);

		//		Data data = new Data();
		//		data.weight = 0.1;
		//		tree.addPoint(data.getDataArray(), data);
		DuelSeed.seed(tree);
	}

	private final LinkedList<Wave<Data>> enemyWaves;
	private final MaximumDeque<State> history;
	private Vector lastEnemyPosition;
	private double lastEnemyEnergy;
	private double lastEnemyVelocity;
	private long timeSinceDirectionChange;
	private double enemyGunHeatx10;

	private int orbitDirection = 1;

	private Wave<Data> lastWave = null;
	private long lastWaveFireTime = 0;

	public Kakeru() {
		enemyWaves = new LinkedList<Wave<Data>>();
		history = new MaximumDeque<State>(MaxTurnHistoryLength);
		lastEnemyPosition = new Vector(-1,-1);
		lastEnemyEnergy = 100;
	}

	@Override
	public void onRoundStarted(Event e) {
		super.onRoundStarted(e);

		/* Set it to be the same as our gun heat on round start! */
		enemyGunHeatx10 = status.getGunHeat() * 10;

		/* Our scans are 1 round late */
		enemyGunHeatx10 -= coolingRate*10;
	}

	private static void prepareWaveData(Wave<Data> wave) {
		wave.data.bulletFlightTime = wave.data.distance / wave.speed;
	}

	@Override
	public void onScannedRobot(ScannedRobotEvent e) {
		super.onScannedRobot(e);

		double angleToEnemy = e.getBearingRadians() + status.getHeadingRadians();
		double enemyVelocity = e.getVelocity();

		State state = new State();

		state.myPosition = myPosition;
		state.enemyPosition = myPosition.projectNew(angleToEnemy, e.getDistance());
		state.data.lateralVelocity = status.getVelocity()*Math.sin(e.getBearingRadians());
		state.data.advancingVelcity = status.getVelocity()*Math.cos(e.getBearingRadians());
		state.data.distance = e.getDistance();
		state.data.globalTime = globalTime;

		int direction = Tools.sign(state.data.lateralVelocity);
		if(direction == orbitDirection) {
			++timeSinceDirectionChange;
		} else {
			timeSinceDirectionChange = 0;
		}
		orbitDirection = direction;

		history.addFirst(state);

		state.data.timeSinceDirectionChange = timeSinceDirectionChange;

		state.data.wallForward = Tools.getWallDistance(
				state.enemyPosition,
				field,
				state.data.distance,
				state.enemyPosition.angleTo(state.myPosition),
				direction);

		state.data.wallReverse = Tools.getWallDistance(
				state.enemyPosition,
				field,
				state.data.distance,
				state.enemyPosition.angleTo(state.myPosition),
				-direction);

		state.data.distLast10 = myPosition.distance(history.get(Tools.min(10,history.size()-1)).myPosition);
		//state.data.distLast20 = myPosition.distance(history.get(Tools.min(20,history.size()-1)).myPosition);
		//state.data.distLast30 = myPosition.distance(history.get(Tools.min(30,history.size()-1)).myPosition);


		if(Math.abs(enemyVelocity) == 0 && lastEnemyVelocity > 2) {

			//Hit the wall!
			lastEnemyEnergy -= Rules.getWallHitDamage(lastEnemyVelocity);
		}

		/*
		 * Fire new imaginary wave if we can, heat waves
		 * are always 2 ticks previous to a real one 'possibly'
		 * being fired
		 */
		if(enemyGunHeatx10 <= 20*coolingRate && enemyGunHeatx10 > 10*coolingRate) {
			double power = 0;
			for(Double d : enemyBulletPower)
				power += d;
			power /= enemyBulletPower.size();

			/* Create the Wave */
			Wave<Data> wave = createMovementWave(power,direction);
			wave.set(simulatedEnemyPosition);
			wave.imaginary = true;

			/*
			 * Only +1 because real waves are -1 in time,
			 * meaning that by the time we detect them, it has
			 * already been 1 turn since it was fired, meaning
			 * heat waves +2 makes it only +1 here. :-)
			 */
			wave.fireTime = time + 1;
			wave.directAngle = wave.angleTo(state.myPosition);

			wave.data = state.data;
			prepareWaveData(wave);

			enemyWaves.add(wave);
		}

		/* Check for new enemy wave */
		double enemyEnergy = e.getEnergy();
		double energyDelta = lastEnemyEnergy - enemyEnergy;
		if(energyDelta > 0 && energyDelta <= 3.0) {
			enemyGunHeatx10 = Rules.getGunHeat(energyDelta)*10;

			State old = history.get(2);
			direction = Tools.sign(old.data.lateralVelocity);

			/* Create the Wave */
			Wave<Data> wave = createMovementWave(energyDelta,direction);
			wave.set(lastEnemyPosition);

			wave.fireTime = time - 1;
			wave.directAngle = wave.angleTo(old.myPosition);

			wave.data = old.data;

			enemyBulletPower.addLast(wave.power);

			boolean add = true;

			if(Math.abs(energyDelta-0.1) < 0.001) {
				/* anti inactivity counter */
				if(time - lastWaveFireTime == 1) {
					enemyWaves.remove(lastWave);
					add = false;
				}

				lastWave = wave;
				lastWaveFireTime = time;
			}

			if(add) {
				prepareWaveData(wave);
				enemyWaves.add(wave);
			}
		}

		lastEnemyPosition = state.enemyPosition;
		lastEnemyVelocity = enemyVelocity;
		lastEnemyEnergy = enemyEnergy;

	}

	private Wave<Data> createMovementWave(double power, int direction) {
		Wave<Data> wave = new Wave<Data>();
		wave.power = power;
		wave.speed = Rules.getBulletSpeed(wave.power);
		wave.escapeAngle = Math.asin(8.0/wave.speed) * direction;
		return wave;
	}

	@Override
	public void onTurnEnded(Event e) {
		super.onTurnEnded(e);

		/* Don't move if we are scanning, this is mostly start
		 * of round fast radar finding protection
		 */
		if(initialTick != 0)
			return;

		/* We track the gun heat because it is important
		 * for our heat/imaginary waves
		 */
		enemyGunHeatx10 -= coolingRate*10;
		if(enemyGunHeatx10 <= 0)
			enemyGunHeatx10 = 0;

		/* Take are of all waves that cannot possibly even conceivably hit us anymore */
		checkWaves();

		/* Run our 'actual' movement */
		doMovement();
	}

	private void checkWaves() {
		Iterator<Wave<Data>> waveIterator = enemyWaves.iterator();
		while(waveIterator.hasNext()) {
			Wave<Data> wave = waveIterator.next();

			double radius = wave.getRadius(time);

			/* /////////////////////// */
			/* DRAW WAVE */
			g.setColor(new Color(128,128,128));
			double escape = Tools.abs(wave.escapeAngle);
			if(wave.imaginary) {
				g.setColor(new Color(255,0,0));
			}
			g.drawArc(
					(int)(wave.x - radius),
					(int)(wave.y - radius),
					(int)(radius*2),
					(int)(radius*2),
					(int)Math.toDegrees(wave.directAngle-escape),
					(int)Math.toDegrees(escape*2));

			/* END DRAW WAVE */
			/* /////////////////////// */

			if(wave.imaginary) {
				/*
				 * if the wave is imaginary (a heat wave)
				 * we should remove it automatically after
				 * 2 ticks which is when we will detect their
				 * actual wave, if they fire
				 */
				if(time - wave.fireTime > 2) {
					waveIterator.remove();
				}
			} else
				/* Clean up waves that can no longer hit us */
				if(wave.didIntersect(myPosition, time)) {
					onWaveRemoval(wave,false);
					waveIterator.remove();
				}
		}
	}

	private Wave<Data> findBestSurfableWave(Wave<Data> exclude) {
		double halfBotWidth = 18 + Math.sin(lastEnemyPosition.angleTo(myPosition))*7.4558441;
		//find the wave soonest to hit us
		Wave<Data> wave = null;
		double bestETA = Double.POSITIVE_INFINITY;
		for(Wave<Data> check : enemyWaves) {
			if(check == exclude)
				continue;

			double distance = myPosition.distance(check) - check.getRadius(time);
			double ETA = distance / check.speed;

			if(distance < 0 && Math.abs(distance) + check.speed > halfBotWidth)
				continue;

			if(ETA < bestETA) {
				wave = check;
				bestETA = ETA;
			}
		}

		return wave;
	}


	private void checkMove(Simulate sim, Wave<Data> wave, int orbitDirection) {
		Move move = predictMove(sim.position,wave,
				sim.heading,sim.velocity,orbitDirection);

		sim.angleToTurn = move.angleToTurn;
		sim.direction = move.direction;
		sim.maxVelocity = Tools.min(sim.maxVelocity, move.maxVelocity);
	}

	/**
	 * Calculate the driving
	 */
	protected Move predictMove(Vector myPosition, Vector orbitCenter,
			double myHeading, double myVelocity, int orbitDirection) {
		Move data = new Move();



		/* Better safe then very very sorry */
		if(orbitDirection == 0)
			orbitDirection = 1;

		double angleToRobot = orbitCenter.angleTo(myPosition);

		/*
		 * if the orbit direction is clockwise/counter, we want to try and
		 * point our robot in that direction, which is why we multiply by it
		 */
		double travelAngle = angleToRobot + (Math.PI/2.0) * orbitDirection;

		/* DONE add distancing to drive method */
		/* TODO add a better distancing method */
		double distance = myPosition.distance(orbitCenter);
		double distancing = ((distance-distBestDistance)/distBestDistance);

		double limit = Tools.limit(
				distMinRotation,
				distMultiplier*distance+distAdder,
				distMaxRotation);

		distancing = Tools.limit(-limit, distancing, limit);

		travelAngle += distancing*orbitDirection;

		travelAngle = Tools.fastSmooth(myPosition, travelAngle, orbitDirection,
				distance, fieldWidth, fieldHeight);

		data.angleToTurn = Utils.normalRelativeAngle(travelAngle - myHeading);
		data.direction = 1;

		/*
		 * If our backend is closer to direction, use that instead, and
		 * inform the caller that we are going to be going in reverse instead
		 */
		if(Tools.abs(data.angleToTurn) > Math.PI/2.0) {
			data.angleToTurn = Utils.normalRelativeAngle(data.angleToTurn - Math.PI);
			data.direction = -1;
		}

		/*
		 * Slow down so we do not ram head long into the walls and can instead turn to avoid them
		 * FIXME We only have to do this because of the wall smoothing, fix that to get rid of this!
		 */
		if(!field.contains(myPosition.projectNew(myHeading, myVelocity*3.25)))
			data.maxVelocity = 0;

		if(!field.contains(myPosition.projectNew(myHeading, myVelocity*5)))
			data.maxVelocity = 4;

		return data;
	}

	private void doMovement() {
		Wave<Data> wave = findBestSurfableWave(null);
		if(wave == null) {
			doWavelessMovement();
			return;
		}

		g.setColor(Color.WHITE);
		g.draw(field.toRectangle2D());

		doSurfingMovement(wave);
	}

	private void doSurfingMovement(Wave<Data> wave) {
		/* Do the actual surfing and such */
		peer.setMaxVelocity(Rules.MAX_VELOCITY);

		Risk[] directions = new Risk[] {
				checkWaveRisk(wave,orbitDirection,0),
				checkWaveRisk(wave,orbitDirection,Rules.MAX_VELOCITY),
				checkWaveRisk(wave,-orbitDirection,Rules.MAX_VELOCITY),
		};

		int bestIndex = 0;
		double minRisk = Double.POSITIVE_INFINITY;
		for(int i = 0; i < directions.length; ++i) {
			if(directions[i].risk < minRisk) {
				bestIndex = i;
				minRisk = directions[i].risk;
			}
		}

		Move move = predictMove(myPosition,wave,
				status.getHeadingRadians(),status.getVelocity(),
				directions[bestIndex].orbitDirection );

		double maxVelocity = Tools.min(directions[bestIndex].maxVelocity,move.maxVelocity);
		peer.setMaxVelocity(maxVelocity);
		peer.setTurnBody(move.angleToTurn);
		peer.setMove(1000*move.direction);
		peer.setCall();
		onMoveDecided(maxVelocity,move.angleToTurn,1000*move.direction);
	}

	protected void onMoveDecided(double maxVelocity, double angleToTurn, double toMove) {

	}

	private Risk checkWaveRisk(Wave<Data> wave, int orbitDirection, double maxVelocity) {
		/*
		 * Simplify the output to our direction chooser
		 */
		Risk waveRisk = new Risk();
		waveRisk.orbitDirection = orbitDirection;
		waveRisk.maxVelocity = maxVelocity;

		/*
		 * Create a simulator for our movement
		 */
		Simulate sim = createSimulator();
		sim.maxVelocity = maxVelocity;
		sim.direction = orbitDirection;

		Simulate core = sim.copy();

		/*
		 * Used for our distance danger
		 */
		double currentDistance = wave.distance(sim.position);
		double predictedDistance = 0;
		int intersected = 0;

		long timeOffset = 0;

		/*
		 * Reset the factors, so we do not get cross risk check contamination
		 */
		wave.resetFactors();

		g.setColor(Color.ORANGE);

		/*
		 * Since our radar is limited to 1200 and the slowest bullet moves at
		 * 11 per second, calculating past 110 is pointlessly impossible, so
		 * it is a safe number to use as our break point.
		 */
		while(timeOffset < 110) {
			//DRAW DANGER MOVEMENT
			g.drawOval((int)sim.position.x-3, (int)sim.position.y-3, 6, 6);

			if(wave.doesIntersect(sim.position, time+timeOffset)) {
				++intersected;
				predictedDistance += wave.distance(sim.position);
			} else if(intersected > 0) {
				predictedDistance /= intersected;
				//predictedDistance = wave.distance(sim.position);

				waveRisk.risk += checkIntersectionRisk(wave);


				Wave<Data> second = findBestSurfableWave(wave);
				if(second != null) {

					//do second wave checking
					Simulate reverse = sim.copy();
					reverse.direction = -orbitDirection;

					double bestRisk = Double.POSITIVE_INFINITY;
					boolean doBreak = false;

					final long timeLimit = timeOffset + 30;

					g.setColor(Color.RED);

					int n0 = 0;
					int n1 = 0;
					while(timeOffset < timeLimit) {
						//for every position along reverse and sim, calculate the PI
						double risk = checkStandingRisk(second,reverse.position);
						if(risk > 0 && risk < bestRisk) {
							bestRisk = risk;
						}

						if(second.intersects(reverse.position, time+timeOffset)) {
							++n0;
						} else if(n0 != 0) {
							doBreak = true;
						}

						risk = checkStandingRisk(second,sim.position);
						if(risk > 0 && risk < bestRisk) {
							bestRisk = risk;
						}

						if(second.intersects(reverse.position, time+timeOffset)) {
							++n1;
						} else if(n1 != 0) {
							if(doBreak) {
								break;
							}
						}

						g.drawRect((int)reverse.position.x-2, (int)reverse.position.y-2, 4, 4);
						g.drawRect((int)sim.position.x-2, (int)sim.position.y-2, 4, 4);

						checkMove(reverse,wave,-orbitDirection);
						reverse.step();
						reverse.maxVelocity = maxVelocity;

						checkMove(sim,wave,orbitDirection);
						sim.step();
						sim.maxVelocity = maxVelocity;

						++timeOffset;
					}

					if(bestRisk < 4) {
						waveRisk.risk += bestRisk/2.0;
						//waveRisk.risk += bestRisk*0.95;
					}
				}

				break;
			}

			checkMove(sim,wave,orbitDirection);
			sim.step();
			sim.maxVelocity = maxVelocity;

			++timeOffset;
		}

		double distanceRisk = currentDistance / predictedDistance;
		distanceRisk *= distanceRisk;
		//		distanceRisk = (distanceRisk + 1.0) / 2.0;

		waveRisk.risk *= distanceRisk;

		return waveRisk;
	}

	private double checkStandingRisk(Wave<Data> wave, Vector target) {
		wave.resetFactors();
		wave.standingIntersection(target);
		return checkIntersectionRisk(wave);
	}

	private double checkIntersectionRisk(Wave<Data> wave) {
		double risk = 0;
		double factorCenter = (wave.minFactor + wave.maxFactor) / 2.0;

		List<Entry<Data>> list = tree.nearestNeighbor(wave.data.getWeightedDataArray(), 24, false);

		for(Entry<Data> e : list) {
			Data data = e.value;

			double eRisk = 0;

			eRisk = 0.1/(1.0+Tools.abs(data.guessfactor-factorCenter));
			if(wave.minFactor < data.guessfactor && wave.maxFactor > data.guessfactor) {
				eRisk += 0.9;
			}

			double timeWeight = AverageSingleRoundTime/(double)(AverageSingleRoundTime+globalTime-data.globalTime);

			double weight = (data.weight / (1.0+e.distance)) * timeWeight;

			risk += eRisk*weight;
		}

		return risk / list.size();
	}

	private void doWavelessMovement() {
		final int initialTurns = (int)Math.ceil(3.0/coolingRate)+4;
		double safeTurns = status.getGunHeat()/coolingRate;

		if(time < initialTurns) {
			/* Do we have enough time to move around before they can start firing? */
			if(safeTurns > TimeBeforeFirstWaveToStartTurning) {
				doMinimalRiskMovement();
			} else {
				Move data = predictMove(myPosition,lastEnemyPosition,
						status.getHeadingRadians(),status.getVelocity(),orbitDirection);
				peer.setTurnBody(data.angleToTurn);
				peer.setMaxVelocity(0);

				peer.setMove(0);
				//peer.setTurnBody(0);
				peer.setMaxVelocity(0);
			}
		} else {
			/*
			 * Things are no longer safe, set our movement to `sane`
			 * 
			 * TODO: RAM if enemy is disabled
			 */
			if(status.getOthers() == 0) {
				/*
				 * No bullets in the air, enemy dead, we can do our victory DANCE (YAY!)
				 */
				doVictoryDance();
			} else {
				//Some other thing to do when nothing is in the air??
				doMinimalRiskMovement();
			}
		}
	}

	protected void doMinimalRiskMovement() {
		double heading = status.getHeadingRadians();
		double velocity = status.getVelocity();

		Rectangle escapeField = new Rectangle(30,30,fieldWidth-60,fieldHeight-60);
		g.setColor(Color.WHITE);
		g.draw(escapeField.toRectangle2D());


		//Do minimal risk movement
		Vector target = myPosition.copy();
		Vector bestTarget = myPosition;
		double angle = 0;
		double bestRisk = checkWavelessRisk(bestTarget);
		double enemyDistance = myPosition.distance(lastEnemyPosition)+18;
		while(angle < Math.PI*2) {
			double targetDistance = Tools.min(200,enemyDistance);

			target.setProject(myPosition, angle, targetDistance);
			if(escapeField.contains(target)) {
				double risk = checkWavelessRisk(target);
				if(risk < bestRisk) {
					bestRisk = risk;
					bestTarget = target.copy();
				}
				g.setColor(Color.BLUE);
				g.drawRect((int)target.x-2, (int)target.y-2, 4, 4);

			}
			angle += Math.PI/32.0;
		}
		g.setColor(bodyColor);
		g.drawRect((int)bestTarget.x-2, (int)bestTarget.y-2, 4, 4);

		double travelAngle = myPosition.angleTo(bestTarget);

		double forward = myPosition.distance(bestTarget);

		double angleToTurn = Utils.normalRelativeAngle(travelAngle - status.getHeadingRadians());
		int direction = 1;

		if(Tools.abs(angleToTurn) > Math.PI/2.0) {
			angleToTurn = Utils.normalRelativeAngle(angleToTurn - Math.PI);
			direction = -1;
		}

		//Slow down so we do not ram head long into the walls and can instead turn to avoid them
		double maxVelocity = Rules.MAX_VELOCITY;

		if(!field.contains(myPosition.projectNew(heading, velocity*3.25)))
			maxVelocity = 0;

		if(!field.contains(myPosition.projectNew(heading, velocity*5)))
			maxVelocity = 4;

		if(angleToTurn > 1.0) {
			maxVelocity = 0;
		}

		peer.setMaxVelocity(maxVelocity);
		peer.setTurnBody(angleToTurn);
		peer.setMove(forward*direction);
	}

	private double checkWavelessRisk(Vector pos) {
		double risk = 0;
		if(lastEnemyEnergy < 20) {
			risk = (lastEnemyEnergy-20)/pos.distanceSq(lastEnemyPosition);
		} else {
			risk = 100.0/pos.distanceSq(lastEnemyPosition);
		}

		for(Line edge : field.getLines()) {
			risk += 5.0/(1.0+edge.ptSegDistSq(pos));
		}

		g.setColor(Color.RED);
		/*
		 * get points between enemy location and corner and add risk!!!!
		 * these are bad places to be! Our hitbox is larger here if nothing else!
		 */
		for(Vector corner : field.getCorners()) {
			corner.add(lastEnemyPosition);
			corner.scale(0.5);
			if(corner.distanceSq(lastEnemyPosition) < 22500 /* 150 */) {
				g.drawRect((int)corner.x-2, (int)corner.y-2, 4, 4);
				risk += 5.0/(1.0+corner.distanceSq(pos));
			}
		}

		return risk;
	}

	/**
	 * Account for our bullets hitting the enemy
	 */
	@Override
	public void onBulletHit(BulletHitEvent e) {
		super.onBulletHit(e);

		/* Get the power of the bullet of ours that hit */
		double bulletPower = e.getBullet().getPower();

		/* Determine how much damage our bullet does to the enemy
		 * and adjust our stored copy to reflect the amount lost
		 */
		lastEnemyEnergy -= Rules.getBulletDamage(bulletPower);
	}

	/**
	 * Account for one of our bullets hitting an enemy bullet.
	 */

	@Override
	public void onBulletHitBullet(BulletHitBulletEvent e) {
		super.onBulletHitBullet(e);

		handleBullet(e.getHitBullet());
	}

	/**
	 * Account for one of the enemies bullets hitting us.
	 */
	@Override
	public void onHitByBullet(HitByBulletEvent e) {
		super.onHitByBullet(e);

		/*
		 * Increase the enemy energy based on how powerful the bullet that
		 * hit us was this is so we can get reliable energy drop detection
		 */
		lastEnemyEnergy += Rules.getBulletHitBonus(e.getPower());

		handleBullet(e.getBullet());
	}

	/**
	 * Handle an actual enemy bullet
	 */
	private void handleBullet(Bullet b) {
		Vector bulletPosition = new Vector(b.getX(),b.getY());

		/* Find the matching wave */
		Iterator<Wave<Data>> wi = enemyWaves.iterator();
		while(wi.hasNext()) {
			Wave<Data> wave = wi.next();

			if(wave.imaginary)
				continue;

			/* check if the power is close enough to be our wave. This
			 * margin is small, only to allow room for rounding errors
			 */
			if(Tools.abs(b.getPower()-wave.power) < 0.01) {
				/* the current distance of the wave */
				double radius = wave.speed*(time - wave.fireTime);
				double radius2 = wave.speed*((time - 1) - wave.fireTime);
				double radius3 = wave.speed*((time - 2) - wave.fireTime);

				double d = wave.distanceSq(bulletPosition);

				double dist = Tools.abs(d-radius*radius);
				double dist2 = Tools.abs(d-radius2*radius2);
				double dist3 = Tools.abs(d-radius3*radius3);

				/* check if the bullets distance from the center of
				 * the wave is close to our waves radius. This margin
				 * is small, only to allow room for rounding errors
				 */
				if(dist < 0.1 || dist2 < 0.1 || dist3 < 0.1) {
					/* FIXME extensively test this detection method */

					/* Alternate method, but why recalculate something? Could be used as an extra check. */
					/* updateTree(wave,wave.angleTo(bulletPosition)); */
					updateTree(wave,b.getHeadingRadians());

					onWaveRemoval(wave,true);

					wi.remove();

					return;
				}
			}
		}

		/* If not found say so, cause something FUCKED UP!!! */
		println("Error: Unknown Bullet Collision");
		println("\tBullet:["+b.getX()+","+b.getY()+"] @ h"+b.getHeadingRadians()+" @ p" + b.getPower());
	}

	protected void onWaveRemoval(Wave<Data> wave, boolean hit) {
		if(hit) {
			printf("HIT@: %.4f %.2f %.2f %.2f %.4f %.4f %.2f %d\n",
					wave.data.guessfactor,
					wave.data.lateralVelocity,
					wave.data.advancingVelcity,
					wave.data.distLast10,
					wave.data.wallForward,
					wave.data.wallReverse,
					wave.data.distance,
					wave.data.timeSinceDirectionChange);
		}
	}

	private void updateTree(Wave<Data> wave, double angleToBullet) {
		double angleOffset =
			Utils.normalRelativeAngle(angleToBullet - wave.directAngle);

		wave.data.guessfactor = angleOffset / wave.escapeAngle;

		//TODO
		//tree.addPoint(wave.data.getDataArray(), wave.data);
		tree.addPoint(wave.data.getWeightedDataArray(), wave.data);
	}
}

/**
 * Recorded State
 */
class State {
	Vector myPosition;
	Vector enemyPosition;
	Data data = new Data();
}

/**
 * Data gained from movement prediction
 */
class Move {
	double angleToTurn = 0;
	double maxVelocity = Rules.MAX_VELOCITY;
	int direction = 1;
}

/**
 * Risk of direction
 */
class Risk {
	double risk;
	double maxVelocity;
	int orbitDirection;
}

